from networkx import Graph

from viscars.recommenders import Recommender


class GraphConvolutionalNetwork(Recommender):

    def __init__(self, graph: Graph, verbose=False):
        super().__init__(graph, verbose)

    def _build_model(self):
        pass

    def predict(self, uid: [], cid: [], **kwargs):
        pass

    def top_n(self, uid: [], cid: [], n: int, **kwargs):
        pass



from enum import Enum

from viscars.recommenders.fastppr import FastPersonalizedPageRank
from viscars.recommenders.pagerank import NetworkXPageRank
from viscars.recommenders.ppr import NetworkXPersonalizedPageRank
from viscars.recommenders.random import RandomRank


class RecommenderType(Enum):
    """Metric types."""
    PAGERANK = 'pagerank'
    PPR = 'ppr'
    FAST_PPR = 'fast_ppr'
    RANDOM = 'random'

    @classmethod
    def reverse_lookup(cls, value):
        """Reverse lookup."""
        for _, member in cls.__members__.items():
            if member.value == value:
                return member
        raise LookupError


class RecommenderFactory:
    """Recommender factory."""
    types_ = {
        RecommenderType.PAGERANK: NetworkXPageRank,
        RecommenderType.PPR: NetworkXPersonalizedPageRank,
        RecommenderType.FAST_PPR: FastPersonalizedPageRank,
        RecommenderType.RANDOM: RandomRank,
    }

    def get(self, type_: RecommenderType):
        """
        Retrieve a recommender.

        :param type_: RecommenderType
        :return: Metric
        """
        cls = self.types_[type_]
        return cls



from fast_pagerank import pagerank_power
import networkx as nx
import numpy as np
from rdflib import Graph
from rdflib.namespace import RDF
from rdflib.extras.external_graph_libs import rdflib_to_networkx_multidigraph

from viscars.namespace import DASHB
from viscars.recommenders import Recommender


class FastPersonalizedPageRank(Recommender):

    def __init__(self, graph: Graph, verbose=False, alpha=0.8, tol=10e-6):
        super().__init__(graph, verbose)

        self.alpha = alpha
        self.tolerance = tol
        self.personalization = None

    def _build_model(self):
        graph_ = Graph()
        graph_ += self.graph.triples((None, None, None))

        self.items = list(graph_.subjects(RDF.type, DASHB.Visualization))

        self.nx_graph = rdflib_to_networkx_multidigraph(graph_).to_undirected()
        self.model = nx.convert_matrix.to_scipy_sparse_matrix(self.nx_graph)

    def set_personalization(self, weight_uid: float = 0.0, weight_cid: float = 0.0):
        weight_others = 1 - weight_uid - weight_cid

        if weight_uid == 0:
            weight_uid = weight_others
        if weight_cid == 0:
            weight_cid = weight_others

        self.personalization = (weight_uid, weight_cid, weight_others)

    def _personalization(self, uid, cid):
        personalization = []

        for node in self.nx_graph.nodes:
            uri = str(node)
            if uri in uid:
                personalization.append(self.personalization[0])
            elif uri in cid:
                personalization.append(self.personalization[1])
            else:
                personalization.append(self.personalization[2])

        return np.array(personalization)

    def run(self, uid: [] = None, cid: [] = None):
        if uid is not None and cid is not None:
            weights = self._personalization(uid, cid)
            return pagerank_power(self.model, p=self.alpha, personalize=weights, tol=self.tolerance)

        return pagerank_power(self.model, p=self.alpha, tol=self.tolerance)

    def predict(self, uid: [] = None, cid: [] = None, **kwargs):
        ranking = self.run(uid, cid)
        pr = {}

        i = 0
        for node in self.nx_graph.nodes:
            pr[node] = ranking[i]
            i += 1

        recommendations = \
            [{'contextId': cid, 'itemId': item, 'score': p} for item, p in pr.items() if item in self.items]
        return sorted(recommendations, key=lambda n: n['score'], reverse=True)

    def top_n(self, uid: [], cid: [], n: int, **kwargs):
        pass



from viscars.data import DataLoader
from viscars.evaluation.metrics import MetricType
from viscars.evaluation.metrics.factory import MetricFactory
from viscars.recommenders import Recommender


class Evaluator:

    def __init__(self, project_id: str, recommender: Recommender, metrics_: [],):
        """
        :param project_id: ID of the project (to load the correct data).
        :param recommender: Recommender
        :param metrics_: List of Metrics
        """
        self.dLoader = DataLoader(project_id)
        self.recommender = recommender
        self.metrics = metrics_

    def evaluate(self):
        raise NotImplementedError



import networkx as nx
from rdflib import Graph
from rdflib.namespace import RDF
from rdflib.extras.external_graph_libs import rdflib_to_networkx_multidigraph

from viscars.namespace import DASHB
from viscars.recommenders import Recommender


class NetworkXPageRank(Recommender):

    def __init__(self, graph: Graph, verbose=False, alpha=0.8, tol=10e-6):
        super().__init__(graph, verbose)

        self.alpha = alpha
        self.tolerance = tol

    def _build_model(self):
        graph_ = Graph()
        graph_ += self.graph.triples((None, None, None))

        self.items = list(graph_.subjects(RDF.type, DASHB.Visualization))
        self.model = rdflib_to_networkx_multidigraph(graph_).to_undirected()

    def set_personalization(self, weight_uid=0, weight_cid=0):
        pass

    def run(self, uid: [] = None, cid: [] = None):
        return nx.pagerank(self.model, alpha=self.alpha, tol=self.tolerance)

    def predict(self, uid: [] = None, cid: [] = None, **kwargs):
        pr = self.run(uid, cid)

        recommendations = \
            [{'contextId': cid, 'itemId': item, 'score': p} for item, p in pr.items() if item in self.items]
        return sorted(recommendations, key=lambda n: n['score'], reverse=True)

    def top_n(self, uid: [], cid: [], n: int, **kwargs):
        pass



from viscars.evaluation.metrics import MetricType
from viscars.evaluation.metrics.metrics import Hits, Precision, Recall, F1, NDCG


class MetricFactory:
    """Evaluation metric factory."""
    types_ = {
        MetricType.HITS: Hits,
        MetricType.PRECISION: Precision,
        MetricType.RECALL: Recall,
        MetricType.F1: F1,
        MetricType.NDCG: NDCG
    }

    def get(self, type_: MetricType, n=None):
        """
        Retrieve a metric.

        :param type_: MetricType
        :param n: Top N
        :return: Metric
        """
        cls = self.types_[type_]
        return cls(n)



from rdflib.term import Literal, URIRef
from rdflib.namespace import DefinedNamespace, Namespace


class DASHB(DefinedNamespace):
    """
    The Dynamic Dashboard Ontology
    """

    _fail = True

    # Classes
    User: URIRef
    UserGroup: URIRef
    Dashboard: URIRef
    Tab: URIRef
    Widget: URIRef
    Property: URIRef
    Metric: URIRef
    Visualization: URIRef
    ObservationBoundary: URIRef

    # User
    memberOf: URIRef

    # Dashboard
    hasTab: URIRef

    # Tab
    hasWidget: URIRef

    # Widget
    createdBy: URIRef
    hasProperty: URIRef
    visualizedBy: URIRef
    hasObservationBoundary: URIRef

    # Property
    produces: URIRef

    # ObservationBoundary
    hasMinBoundary: Literal
    hasMaxBoundary: Literal
    hasBoundaryLabel: Literal

    _NS = Namespace('https://docs.dynamicdashboard.ilabt.imec.be/ontologies/dashboard#')



"""Evaluation metrics."""
from abc import ABC, abstractmethod
from enum import Enum


class MetricType(Enum):
    """Metric types."""
    HITS = 'hits'
    PRECISION = 'precision'
    RECALL = 'recall'
    F1 = 'f1'
    NDCG = 'ndcg'

    @classmethod
    def reverse_lookup(cls, value):
        """Reverse lookup."""
        for _, member in cls.__members__.items():
            if member.value == value:
                return member
        raise LookupError


class Metric(ABC):
    type_ = None

    def __init__(self, n=None):
        self.n = n

    @abstractmethod
    def calculate(self, ranking, truth):
        pass

    def __str__(self):
        name = self.type_.value
        return f'{name}@{self.n}' if self.n is not None else name


import random
from rdflib import Graph
from rdflib.namespace import RDF
from viscars.namespace import DASHB
from viscars.recommenders import Recommender


class RandomRank(Recommender):

    def __init__(self, graph: Graph, verbose=False):
        super().__init__(graph, verbose)

    def _build_model(self):
        graph_ = Graph()
        graph_ += self.graph.triples((None, None, None))

        self.items = list(graph_.subjects(RDF.type, DASHB.Visualization))

    def run(self, uid: [] = None, cid: [] = None):
        ranking = {}

        for i in self.items:
            ranking[i] = random.randint(0, 1)
        return ranking

    def predict(self, uid: [] = None, cid: [] = None, **kwargs):
        pr = self.run(uid, cid)

        recommendations = \
            [{'contextId': cid, 'itemId': item, 'score': p} for item, p in pr.items() if item in self.items]
        return sorted(recommendations, key=lambda n: n['score'], reverse=True)

    def top_n(self, uid: [], cid: [], n: int, **kwargs):
        pass



from viscars.evaluation.evaluators import Evaluator
from viscars.recommenders import Recommender


class LeaveOneOutCrossValidation(Evaluator):

    def __init__(self, project_id: str, recommender: Recommender, metrics: []):
        """
        :param recommender: Recommender
        :param metrics: List of MetricTypes
        :param k: Number of folds
        """
        super().__init__(project_id, recommender, metrics)

    def _split_ratings(self, uid):
        ratings = self.dLoader.ratings
        train = ratings[ratings['user'] != uid]
        test = ratings[ratings['user'] == uid]

        return train, test

    def evaluate(self, **kwargs):
        users = self.dLoader.user_metadata

        result = {'folds': [], 'result': {}}
        for uid in users['id'].unique():
            train, test = self._split_ratings(uid)

            graph = self.dLoader.build_subgraph_from_ratings(train)
            self.recommender.set_graph(graph)

            fold_scores = {}


            df_user = test.loc[test['user'] == uid]

            for cid in df_user['context']:
                predictions = self.recommender.predict(uid, cid, **kwargs)
                recommendations = [r['itemId'] for r in predictions]

                truth = []
                t_user = test.loc[test['user'] == uid]
                for idx, row in t_user.iterrows():
                    if row['context'] == cid:
                        truth.append(row['item'])
                #truth = list(test.loc[test['user'] == uid].loc[test['context'] == cid]['item'])

                for metric in self.metrics:
                    if str(metric) not in fold_scores.keys():
                        fold_scores[str(metric)] = []
                    score = metric.calculate(recommendations, truth)
                    fold_scores[str(metric)].append(score)

            result_for_fold = {'uid': uid}
            for metric_type, scores in fold_scores.items():
                avg = sum(scores) / len(scores)
                result_for_fold[metric_type] = avg

                if metric_type not in result['result'].keys():
                    result['result'][metric_type] = []
                result['result'][metric_type].append(avg)
            result['folds'].append(result_for_fold)

        final_results = {}
        for metric_type, score in result['result'].items():
            final_results[metric_type] = sum(score) / len(score)

        result['result'] = final_results
        return result



from sklearn.model_selection import KFold

from viscars.evaluation.evaluators import Evaluator
from viscars.recommenders import Recommender


class KFoldCrossValidation(Evaluator):

    def __init__(self, project_id: str, recommender: Recommender, metrics: [], k=5):
        """
        :param project_id: ID of the project (to load the correct data).
        :param recommender: Recommender
        :param metrics: List of Metrics
        :param k: Number of folds
        """
        super().__init__(project_id, recommender, metrics)

        self.k = k

    def evaluate(self, **kwargs):
        ratings = self.dLoader.ratings
        kf = KFold(n_splits=self.k, shuffle=True)

        n_fold = 0

        result = {'folds': [], 'result': {}}
        for train_idx, test_idx in kf.split(ratings):
            train = ratings.iloc[train_idx]
            test = ratings.iloc[test_idx]

            graph = self.dLoader.build_subgraph_from_ratings(train)
            self.recommender.set_graph(graph)

            fold_scores = {}

            for uid in test['user'].unique():
                df_user = test.loc[test['user'] == uid]

                for cid in df_user['context']:
                    predictions = self.recommender.predict(uid, cid, **kwargs)
                    recommendations = [r['itemId'] for r in predictions]

                    truth = []
                    t_user = test.loc[test['user'] == uid]
                    for idx, row in t_user.iterrows():
                        if row['context'] == cid:
                            truth.append(row['item'])
                    # truth = list(test.loc[test['user'] == uid].loc[test['context'] == cid]['item'])

                    for metric in self.metrics:
                        if str(metric) not in fold_scores.keys():
                            fold_scores[str(metric)] = []
                        score = metric.calculate(recommendations, truth)
                        fold_scores[str(metric)].append(score)

            result_for_fold = {}
            for metric, scores in fold_scores.items():
                avg = sum(scores) / len(scores)
                result_for_fold[metric] = avg

                if metric not in result['result'].keys():
                    result['result'][metric] = []
                result['result'][metric].append(avg)
            result['folds'].append(result_for_fold)

            n_fold += 1

        final_results = {}
        for metric_type, score in result['result'].items():
            final_results[metric_type] = sum(score) / len(score)

        result['result'] = final_results
        return result



from abc import ABC, abstractmethod
from rdflib import Graph


class Recommender(ABC):

    def __init__(self, graph: Graph, verbose=False, **kwargs):
        self.graph = graph
        self.verbose = verbose

        self.items = []

        self._build_model()

    @abstractmethod
    def _build_model(self):
        pass

    @abstractmethod
    def predict(self, uid: [], cid: [], **kwargs):
        """

        :param uid: User ID
        :param cid: List of Context IDs
        :param n: Number of recommendations to return.
        :return: Prediction value
        """
        pass

    @abstractmethod
    def top_n(self, uid: [], cid: [], n: int, **kwargs):
        pass

    def get_graph(self) -> Graph:
        return self.graph

    def set_graph(self, graph: Graph):
        self.graph = graph
        self._build_model()



import pandas as pd
from rdflib import Graph

from viscars.namespace import DASHB
from viscars.utils import clean_graph


class DataLoader:

    def __init__(self, graph: Graph):
        self.graph = graph
        clean_graph(graph)

        self.ratings = self.extract_ratings_from_graph()
        self.user_metadata = self.extract_users_from_graph()
        self.context_metadata = self.extract_contexts_from_graph()
        self.item_metadata = self.extract_items_from_graph()

    # def load_graph(self) -> Graph:
    #     graph = Graph()
    #     graph.parse(f'./data/{self.project}/graph.ttl', format='n3')
    #
    #     return clean_graph(graph)

    def build_subgraph_from_ratings(self, ratings: pd.DataFrame) -> Graph:
        sub_graph = Graph()

        for idx, row in self.context_metadata.iterrows():
            cid = row['id']
            sub_graph += self.graph.triples((cid, None, None))

        for idx, row in self.user_metadata.iterrows():
            uid = row['id']
            sub_graph += self.graph.triples((uid, DASHB.memberOf, None))

        for idx, row in self.item_metadata.iterrows():
            iid = row['id']
            sub_graph += self.graph.triples((iid, None, None))

        for idx, row in ratings.iterrows():
            uid = row['user']
            iid = row['item']
            cid = row['context']

            sub_graph += self.graph.triples((uid, None, None))
            sub_graph += self.graph.triples((None, None, uid))
            sub_graph += self.graph.triples((iid, None, None))
            sub_graph += self.graph.triples((None, None, iid))

            for cid_ in cid:
                sub_graph += self.graph.triples((cid_, None, None))
                sub_graph += self.graph.triples((None, None, cid_))

        return sub_graph

    def extract_ratings_from_graph(self) -> pd.DataFrame:
        qry = '''
            prefix dashb: <https://docs.dynamicdashboard.ilabt.imec.be/ontologies/dashboard#>

            SELECT ?widget ?user ?visualization WHERE {
                ?widget dashb:visualizedBy ?visualization ;
                        dashb:createdBy ?user .
            }
        '''
        result = self.graph.query(qry)

        ratings = {'user': [], 'item': [], 'rating': [], 'context': []}
        for row in result:
            widget_ = row[0]
            user_ = row[1]

            qry = f'''
                prefix dashb: <https://docs.dynamicdashboard.ilabt.imec.be/ontologies/dashboard#>

                SELECT ?property WHERE {{
                    <{widget_}> dashb:hasProperty ?property .
                }}
            '''
            properties = [str(row[0]) for row in self.graph.query(qry)]

            ratings.get('user').append(user_)
            ratings.get('item').append(row[2])
            ratings.get('rating').append(5.0)
            ratings.get('context').append(properties)
        ratings_df = pd.DataFrame.from_dict(ratings)

        return ratings_df

    def extract_users_from_graph(self) -> pd.DataFrame:
        qry = '''
            prefix dashb: <https://docs.dynamicdashboard.ilabt.imec.be/ontologies/dashboard#>

            SELECT ?user ?username ?role WHERE {
                ?user dashb:memberOf ?role .
            }
        '''
        user_metadata = {'id': [], 'type': []}

        result = self.graph.query(qry)
        for row in result:
            user_metadata.get('id').append(row[0])
            user_metadata.get('type').append(row[1])
        user_metadata_df = pd.DataFrame.from_dict(user_metadata)

        return user_metadata_df

    def extract_contexts_from_graph(self) -> pd.DataFrame:
        qry = '''
            prefix dashb: <https://docs.dynamicdashboard.ilabt.imec.be/ontologies/dashboard#>

            SELECT ?property ?metric WHERE {
                ?property dashb:produces ?metric .
            }
        '''
        context_metadata = {'id': [], 'type': []}

        result = self.graph.query(qry)
        for row in result:
            context_metadata.get('id').append(row[0])
            context_metadata.get('type').append(row[1])
        context_metadata_df = pd.DataFrame.from_dict(context_metadata)

        return context_metadata_df

    def extract_items_from_graph(self) -> pd.DataFrame:
        qry = '''
            prefix dashb: <https://docs.dynamicdashboard.ilabt.imec.be/ontologies/dashboard#>

            SELECT ?visualization WHERE {
                ?visualization a dashb:Visualization .
            }
        '''
        item_metadata = {'id': []}

        result = self.graph.query(qry)
        for row in result:
            item_metadata.get('id').append(row[0])
        item_metadata_df = pd.DataFrame.from_dict(item_metadata)

        return item_metadata_df

    def get_graph(self) -> Graph:
        return self.graph

    def set_graph(self, graph: Graph):
        self.graph = graph

        self.ratings = self.extract_ratings_from_graph()
        self.user_metadata = self.extract_users_from_graph()
        self.context_metadata = self.extract_contexts_from_graph()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['viscars',
 'viscars.data',
 'viscars.evaluation.evaluators',
 'viscars.evaluation.metrics',
 'viscars.namespace',
 'viscars.recommenders',
 'viscars.training',
 'viscars.utils']

package_data = \
{'': ['*']}

install_requires = \
['fast-pagerank>=0.0.4,<0.0.5',
 'matplotlib>=3.4.2,<4.0.0',
 'networkx>=2.6.3,<3.0.0',
 'numpy>=1.20.2,<2.0.0',
 'pandas>=1.2.4,<2.0.0',
 'rdflib>=6.0.2,<7.0.0']

setup_kwargs = {
    'name': 'viscars',
    'version': '0.0.10',
    'description': 'VisCARS: Knowledge Graph-based Context-Aware Recommender System for Visualizations',
    'long_description': '# VisCARS: Graph-Based Context-Aware Visualization Recommendation System\n\n![version](https://img.shields.io/pypi/v/viscars)\n\n## Installation\n\n```\npip install viscars\n```\n\n## Citation\n',
    'author': 'Pieter Moens',
    'author_email': 'pieter.moens@ugent.be',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/predict-idlab/VisCARS',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import networkx as nx
import numpy as np
from rdflib import Graph
from rdflib.namespace import RDF
from rdflib.extras.external_graph_libs import rdflib_to_networkx_multidigraph

from viscars.namespace import DASHB
from viscars.recommenders import Recommender


class NetworkXPersonalizedPageRank(Recommender):

    def __init__(self, graph: Graph, verbose=False, alpha=0.8, tol=10e-6):
        super().__init__(graph, verbose)

        self.alpha = alpha
        self.tolerance = tol
        self.personalization = None

    def _build_model(self):
        graph_ = Graph()
        graph_ += self.graph.triples((None, None, None))

        self.items = list(graph_.subjects(RDF.type, DASHB.Visualization))
        self.model = rdflib_to_networkx_multidigraph(graph_).to_undirected()

    def set_personalization(self, weight_uid=0, weight_cid=0):
        weight_others = 1 - weight_uid - weight_cid

        if weight_uid == 0:
            weight_uid = weight_others
        if weight_cid == 0:
            weight_cid = weight_others

        self.personalization = (weight_uid, weight_cid, weight_others)

    def _personalization(self, uid, cid):
        personalization = {}

        for node in self.model.nodes:
            uri = str(node)
            if uri in uid:
                personalization[node] = self.personalization[0]
            elif uri in cid:
                personalization[node] = self.personalization[1]
            else:
                personalization[node] = self.personalization[2]

        return personalization

    def run(self, uid: [] = None, cid: [] = None):
        if uid is not None and cid is not None:  # TODO: Check if UID and CID exist in graph -> ZeroDivisionError
            weights = self._personalization(uid, cid)
            return nx.pagerank(self.model, alpha=self.alpha, personalization=weights,
                               tol=self.tolerance)

        return nx.pagerank(self.model, alpha=self.alpha, tol=self.tolerance)

    def predict(self, uid: [] = None, cid: [] = None, **kwargs):
        pr = self.run(uid, cid)

        recommendations = \
            [{'contextId': cid, 'itemId': item, 'score': p} for item, p in pr.items()
             if item in self.items]
        return sorted(recommendations, key=lambda n: n['score'], reverse=True)

    def top_n(self, uid: [], cid: [], n: int, **kwargs):
        pass



from matplotlib import pyplot as plt
import networkx as nx
from rdflib import Graph
from rdflib.namespace import RDF, SOSA, URIRef
from rdflib.extras.external_graph_libs import rdflib_to_networkx_multidigraph

from viscars.namespace import DASHB

EXPECTED_PROPERTIES = [DASHB.hasProperty, DASHB.visualizedBy, DASHB.createdBy, DASHB.memberOf,
                       DASHB.produces]


def visualize_graph(graph: Graph, colored=False, save=False):
    graph_ = Graph()
    graph_ += graph.triples((None, None, None))

    users = list(graph_.subjects(RDF.type, DASHB.User))
    items = list(graph_.subjects(RDF.type, DASHB.Visualization))
    contexts = list(graph_.subjects(RDF.type, SOSA.ObservableProperty))

    graph_.remove((None, RDF.type, None))

    G = rdflib_to_networkx_multidigraph(graph_).to_undirected()

    colors = []
    for node in G.nodes():
        if URIRef(node) in users:
            colors.append('red')
        elif URIRef(node) in items:
            colors.append('green')
        elif URIRef(node) in contexts:
            colors.append('#1f78b4')
        else:
            colors.append('black')

    # Plot Networkx instance of RDF Graph
    fig = plt.figure(figsize=(16, 16))
    ax = fig.add_subplot(111)

    colors = colors if colored else None
    nx.draw(G, with_labels=False, node_color=colors, ax=ax)

    if save:
        plt.savefig('graph.png', dpi=600)
    plt.show()


def clean_graph(graph: Graph):
    G = Graph()

    for ep in EXPECTED_PROPERTIES:
        G += graph.triples((None, ep, None))

    G += graph.triples((None, RDF.type, None))

    return G



import math

from viscars.evaluation.metrics import Metric, MetricType


class Hits(Metric):
    """"Hits or True Positives."""
    type_ = MetricType.HITS

    def calculate(self, ranking, truth):
        hits = 0

        for idx, item in enumerate(ranking):
            if self.n is not None and idx >= self.n:
                break

            if item in truth:
                hits += 1

        return hits


class Precision(Metric):
    """"Precision = TP / (TP + FP)."""
    type_ = MetricType.PRECISION

    def calculate(self, ranking, truth):
        hits = Hits(self.n)

        if self.n is not None:
            ranking = ranking[:self.n]

        return hits.calculate(ranking, truth) / len(ranking)


class Recall(Metric):
    """"Recall = TP / (TP + FN)."""
    type_ = MetricType.RECALL

    def calculate(self, ranking, truth):
        hits = Hits()

        if self.n is not None:
            ranking = ranking[:self.n]

        if len(truth) == 0:
            return 0
        return hits.calculate(ranking, truth) / len(truth)


class F1(Metric):
    """"F1 = 2 * Precision * Recall / (Precision + Recall)."""
    type_ = MetricType.F1

    def calculate(self, ranking, truth):
        if self.n is not None:
            ranking = ranking[:self.n]

        precision = Precision().calculate(ranking, truth)
        recall = Recall().calculate(ranking, truth)

        if precision + recall == 0:
            return 0
        return 2 * precision * recall / (precision + recall)


class NDCG(Metric):
    """"
    Compute Normalized Discounted Cumulative Gain.

    Sum the true scores ranked in the order induced by the predicted scores, after applying a logarithmic discount.
    Then divide by the best possible score (Ideal DCG, obtained for a perfect ranking) to obtain a score between 0 and 1.
    """
    type_ = MetricType.NDCG

    def calculate(self, ranking, truth):
        dcg = 0
        idcg = self._idcg(len(truth))

        if self.n is not None:
            ranking = ranking[:self.n]

        for idx, item in enumerate(ranking):
            if item in truth:
                dcg += 1 / math.log2(idx + 2)

        if idcg == 0:
            return 0
        return dcg / idcg

    @staticmethod
    def _idcg(n):
        idcg = 0

        for i in range(n):
            idcg += 1 / math.log2(i + 2)

        return idcg
